@@ -2,19 +2,29 @@
 
 from __future__ import annotations
 
-import asyncio
+from asyncio import (
+    ensure_future,
+    get_event_loop,
+    sleep,
+    wait_for,
+    Future,
+    Queue,
+    QueueEmpty,
+    Protocol,
+    Transport,
+)
+from asyncio.exceptions import InvalidStateError, TimeoutError
 
 from collections.abc import Callable
 from logging import Logger
 from typing import Any
 
-from .const import Action, FunctionalDomain
-from .packet import decode_packet
+from .const import Action, FunctionalDomain, QUEUE_FREQUENCY
+from .packet import NackPacket, Packet
 from .socket_client import SocketClient
-from .utils import encode_temperature, generate_command_bytes
 
 
-class _AprilaireClientProtocol(asyncio.Protocol):
+class _AprilaireClientProtocol(Protocol):
     """Protocol for interacting with the thermostat over socket connection"""
 
     def __init__(
@@ -28,112 +38,122 @@ class _AprilaireClientProtocol(asyncio.Protocol):
         self.reconnect_action = reconnect_action
         self.logger = logger
 
-        self.transport: asyncio.Transport = None
+        self.transport: Transport = None
 
-        self.command_queue = asyncio.Queue()
+        self.packet_queue = Queue()
 
-        self.sequence = 1
+        self.sequence = 0
 
-    async def _send_command(
-        self,
-        action: Action,
-        functional_domain: FunctionalDomain,
-        attribute: int,
-        extra_payload: list[int] = None,
-    ) -> None:
+    def _get_sequence(self):
+        self.sequence = (self.sequence + 1) % 128
+
+        return self.sequence
+
+    async def _send_packet(self, packet: Packet) -> None:
         """Send a command to the thermostat"""
-        command_bytes = generate_command_bytes(
-            self.sequence,
-            action,
-            functional_domain,
-            attribute,
-            extra_payload=extra_payload,
-        )
+
+        packet.sequence = self._get_sequence()
 
         self.logger.debug(
             "Queuing data, sequence=%d, action=%s, functional_domain=%s, attribute=%d",
-            self.sequence,
-            str(action),
-            str(functional_domain),
-            attribute,
+            packet.sequence,
+            str(packet.action),
+            str(packet.functional_domain),
+            packet.attribute,
         )
 
-        self.sequence = (self.sequence + 1) % 128
+        await self.packet_queue.put(packet)
 
-        await self.command_queue.put(command_bytes)
-
-    def _empty_command_queue(self):
+    def _empty_packet_queue(self):
         try:
-            for _ in range(self.command_queue.qsize()):
-                self.command_queue.get_nowait()
-                self.command_queue.task_done()
+            for _ in range(self.packet_queue.qsize()):
+                self.packet_queue.get_nowait()
+                self.packet_queue.task_done()
         except:  # pylint: disable=bare-except
             pass
 
-    async def _process_command_queue(self):
+    async def _queue_loop(self):
+        """Periodically send items from the queue"""
         while True:
-            command_bytes = await self.command_queue.get()
+            try:
+                packet: Packet
 
-            if not command_bytes:
-                continue
+                while packet := self.packet_queue.get_nowait():
+                    if self.transport:
+                        serialized_packet = packet.serialize()
 
-            if not self.transport:
-                break
+                        self.logger.info("Sent data: %s", serialized_packet.hex(" "))
 
-            self.logger.debug("Sending data %s", command_bytes.hex(" ", 1))
+                        self.transport.write(serialized_packet)
+            except QueueEmpty:
+                pass
 
-            self.transport.write(command_bytes)
+            await sleep(QUEUE_FREQUENCY)
 
-    def connection_made(self, transport: asyncio.Transport):
+    def connection_made(self, transport: Transport):
         """Called when a connection has been made to the socket"""
         self.logger.info("Aprilaire connection made")
 
         self.transport = transport
-        self._empty_command_queue()
+        self._empty_packet_queue()
 
-        asyncio.ensure_future(self._process_command_queue())
+        ensure_future(self._queue_loop())
 
         async def _update_status():
-            await asyncio.sleep(2)
+            await sleep(2)
 
             await self.read_mac_address()
             await self.read_thermostat_status()
+            await self.read_control()
             await self.read_sensors()
+            await self.read_thermostat_name()
             await self.configure_cos()
             await self.sync()
 
-        asyncio.ensure_future(_update_status())
+        ensure_future(_update_status())
 
     def data_received(self, data: bytes) -> None:
         """Called when data has been received from the socket"""
-        self.logger.info("Aprilaire data received")
+        self.logger.info("Aprilaire data received %s", data.hex(" "))
 
-        decoded_packets = decode_packet(data)
-
-        for decoded_packet in decoded_packets:
-            if "event" not in decoded_packet:
-                self.logger.warning("Event data missing")
-                return
-
-            (action, functional_domain, attribute) = decoded_packet["event"]
+        parsed_packets = Packet.parse(data)
 
+        for packet in parsed_packets:
             self.logger.debug(
                 "Received data, action=%s, functional_domain=%s, attribute=%d",
-                action,
-                functional_domain,
-                attribute,
+                str(packet.action),
+                str(packet.functional_domain),
+                packet.attribute,
             )
 
-            if "error" in decoded_packet:
-                error = decoded_packet["error"]
+            if isinstance(packet, NackPacket):
+                self.logger.error(
+                    "Received NACK for attribute %d", packet.nack_attribute
+                )
+                continue
+
+            if "error" in packet.data:
+                error = packet.data["error"]
 
                 if error != 0:
                     self.logger.error("Thermostat error: %d", error)
 
+            if (
+                packet.action == Action.COS
+                and packet.functional_domain == FunctionalDomain.CONTROL
+                and packet.attribute == 1
+                and packet.data.get("mode") == 1
+            ):
+                self.logger.info("Re-reading control because of COS with mode==1")
+
+                ensure_future(self.read_control())
+
+                continue
+
             if self.data_received_callback:
-                asyncio.ensure_future(
+                ensure_future(
                     self.data_received_callback(
-                        functional_domain, attribute, decoded_packet
+                        packet.functional_domain, packet.attribute, packet.data
                     )
                 )
 
@@ -142,161 +162,163 @@ class _AprilaireClientProtocol(asyncio.Protocol):
         self.logger.info("Aprilaire connection lost")
 
         if self.data_received_callback:
-            asyncio.ensure_future(
+            ensure_future(
                 self.data_received_callback(
                     FunctionalDomain.NONE, 0, {"available": False}
                 )
             )
 
+        self.transport = None
+
         if self.reconnect_action:
-            asyncio.ensure_future(self.reconnect_action())
+            ensure_future(self.reconnect_action())
 
     async def read_sensors(self):
         """Send a request for updated sensor data"""
-        await self._send_command(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)
+        await self._send_packet(
+            Packet(Action.READ_REQUEST, FunctionalDomain.SENSORS, 2)
+        )
 
     async def read_control(self):
         """Send a request for updated control data"""
-        await self._send_command(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)
+        await self._send_packet(
+            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 1)
+        )
 
     async def read_scheduling(self):
         """Send a request for updated scheduling data"""
-        await self._send_command(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)
+        await self._send_packet(
+            Packet(Action.READ_REQUEST, FunctionalDomain.SCHEDULING, 4)
+        )
 
     async def update_mode(self, mode: int):
         """Send a request to update the mode"""
-        await self._send_command(
-            Action.WRITE,
-            FunctionalDomain.CONTROL,
-            1,
-            extra_payload=[
-                mode,  # Mode
-                0,  # Fan Mode (0 to not set)
-                0,  # Heat Setpoint (0 to not set)
-                0,  # Cool Setpoint (0 to not set)
-            ],
+        await self._send_packet(
+            Packet(
+                Action.WRITE,
+                FunctionalDomain.CONTROL,
+                1,
+                data={
+                    "mode": mode,
+                    "fan_mode": 0,
+                    "heat_setpoint": 0,
+                    "cool_setpoint": 0,
+                },
+            )
         )
 
     async def update_fan_mode(self, fan_mode: int):
         """Send a request to update the fan mode"""
-        await self._send_command(
-            Action.WRITE,
-            FunctionalDomain.CONTROL,
-            1,
-            extra_payload=[
-                0,  # Mode (0 to not set)
-                fan_mode,  # Fan Mode
-                0,  # Heat Setpoint (0 to not set)
-                0,  # Cool Setpoint (0 to not set)
-            ],
+        await self._send_packet(
+            Packet(
+                Action.WRITE,
+                FunctionalDomain.CONTROL,
+                1,
+                data={
+                    "mode": 0,
+                    "fan_mode": fan_mode,
+                    "heat_setpoint": 0,
+                    "cool_setpoint": 0,
+                },
+            )
         )
 
     async def update_setpoint(self, cool_setpoint: int, heat_setpoint: int):
         """Send a request to update the setpoint"""
-        await self._send_command(
-            Action.WRITE,
-            FunctionalDomain.CONTROL,
-            1,
-            extra_payload=[
-                0,  # Mode
-                0,  # Fan
-                encode_temperature(heat_setpoint),  # Heat Setpoint
-                encode_temperature(cool_setpoint),  # Cool Setpoint
-            ],
+        await self._send_packet(
+            Packet(
+                Action.WRITE,
+                FunctionalDomain.CONTROL,
+                1,
+                data={
+                    "mode": 0,
+                    "fan_mode": 0,
+                    "heat_setpoint": heat_setpoint,
+                    "cool_setpoint": cool_setpoint,
+                },
+            )
         )
 
     async def set_hold(self, hold: int):
         """Send a request to set the hold status"""
-        await self._send_command(
-            Action.WRITE,
-            FunctionalDomain.SCHEDULING,
-            4,
-            extra_payload=[
-                hold,  # Hold
-                0,  # Fan
-                0,  # Heat Setpoint
-                0,  # Cool Setpoint
-                0,  # DEH Vacation
-                0,  # End Minute
-                0,  # End Hour
-                0,  # End Date
-                0,  # End Month
-                0,  # End Year
-            ],
+
+        await self._send_packet(
+            Packet(
+                Action.WRITE,
+                FunctionalDomain.SCHEDULING,
+                4,
+                data={"hold": hold},
+            )
         )
 
     async def sync(self):
         """Send a request to sync data"""
-        await self._send_command(
-            Action.WRITE,
-            FunctionalDomain.STATUS,
-            2,
-            extra_payload=[
-                1,  # Sync
-            ],
+        await self._send_packet(
+            Packet(
+                Action.WRITE,
+                FunctionalDomain.STATUS,
+                2,
+                data={"synced": 1},
+            )
         )
 
     async def configure_cos(self):
         """Send a request to configure the COS settings"""
-        await self._send_command(
-            Action.WRITE,
-            FunctionalDomain.STATUS,
-            1,
-            extra_payload=[
-                1,  # Installer Thermostat Settings
-                0,  # Contractor Information
-                0,  # Air Cleaning Installer Variable
-                0,  # Humidity Control Installer Settings
-                0,  # Fresh Air Installer Settings
-                1,  # Thermostat Setpoint & Mode Settings
-                0,  # Dehumidification Setpoint
-                0,  # Humidification Setpoint
-                0,  # Fresh Air Setting
-                0,  # Air Cleaning Settings
-                1,  # Thermostat IAQ Available
-                0,  # Schedule Settings
-                0,  # Away Settings
-                0,  # Schedule Day
-                1,  # Schedule Hold
-                0,  # Heat Blast
-                0,  # Service Reminders Status
-                0,  # Alerts Status
-                0,  # Alerts Settings
-                0,  # Backlight Settings
-                1,  # Thermostat Location & Name
-                0,  # Reserved
-                1,  # Controlling Sensor Values
-                0,  # Over the air ODT update timeout
-                1,  # Thermostat Status
-                1,  # IAQ Status
-                1,  # Model & Revision
-                0,  # Support Module
-                0,  # Lockouts
-            ],
+        await self._send_packet(
+            Packet(
+                Action.WRITE,
+                FunctionalDomain.STATUS,
+                1,
+                raw_data=[
+                    1,  # Installer Thermostat Settings
+                    0,  # Contractor Information
+                    0,  # Air Cleaning Installer Variable
+                    0,  # Humidity Control Installer Settings
+                    0,  # Fresh Air Installer Settings
+                    1,  # Thermostat Setpoint & Mode Settings
+                    0,  # Dehumidification Setpoint
+                    0,  # Humidification Setpoint
+                    0,  # Fresh Air Setting
+                    0,  # Air Cleaning Settings
+                    1,  # Thermostat IAQ Available
+                    0,  # Schedule Settings
+                    0,  # Away Settings
+                    0,  # Schedule Day
+                    1,  # Schedule Hold
+                    0,  # Heat Blast
+                    0,  # Service Reminders Status
+                    0,  # Alerts Status
+                    0,  # Alerts Settings
+                    0,  # Backlight Settings
+                    1,  # Thermostat Location & Name
+                    0,  # Reserved
+                    1,  # Controlling Sensor Values
+                    0,  # Over the air ODT update timeout
+                    1,  # Thermostat Status
+                    1,  # IAQ Status
+                    1,  # Model & Revision
+                    0,  # Support Module
+                    0,  # Lockouts
+                ],
+            )
         )
 
     async def read_mac_address(self):
         """Send a request to get identification data (including MAC address)"""
-        await self._send_command(
-            Action.READ_REQUEST,
-            FunctionalDomain.IDENTIFICATION,
-            2,
+        await self._send_packet(
+            Packet(Action.READ_REQUEST, FunctionalDomain.IDENTIFICATION, 2)
         )
 
     async def read_thermostat_status(self):
         """Send a request for thermostat status"""
-        await self._send_command(
-            Action.READ_REQUEST,
-            FunctionalDomain.CONTROL,
-            7,
+        await self._send_packet(
+            Packet(Action.READ_REQUEST, FunctionalDomain.CONTROL, 7)
         )
 
     async def read_thermostat_name(self):
         """Send a reques for the thermostat name"""
-        await self._send_command(
-            Action.READ_REQUEST,
-            FunctionalDomain.IDENTIFICATION,
-            5,
+        await self._send_packet(
+            Packet(Action.READ_REQUEST, FunctionalDomain.IDENTIFICATION, 5)
         )
 
 
@@ -312,6 +334,8 @@ class AprilaireClient(SocketClient):
         reconnect_interval: int = None,
         retry_connection_interval: int = None,
     ) -> None:
+        self.protocol: _AprilaireClientProtocol = None
+
         super().__init__(
             host,
             port,
@@ -321,11 +345,14 @@ class AprilaireClient(SocketClient):
             retry_connection_interval,
         )
 
-        self.futures: dict[tuple[FunctionalDomain, int], list[asyncio.Future]] = {}
+        self.futures: dict[tuple[FunctionalDomain, int], list[Future]] = {}
+
+    async def _reconnect_with_delay(self):
+        await super()._reconnect(self.retry_connection_interval)
 
     def create_protocol(self):
         return _AprilaireClientProtocol(
-            self.data_received, self._reconnect, self.logger
+            self.data_received, self._reconnect_with_delay, self.logger
         )
 
     async def data_received(
@@ -344,7 +371,7 @@ class AprilaireClient(SocketClient):
         for future in futures_to_complete:
             try:
                 future.set_result(data)
-            except asyncio.exceptions.InvalidStateError:
+            except InvalidStateError:
                 pass
 
     def state_changed(self):
@@ -362,7 +389,7 @@ class AprilaireClient(SocketClient):
     ):
         """Wait for a response for a particular request"""
 
-        loop = asyncio.get_event_loop()
+        loop = get_event_loop()
         future = loop.create_future()
 
         future_key = (functional_domain, attribute)
@@ -373,8 +400,8 @@ class AprilaireClient(SocketClient):
         self.futures[future_key].append(future)
 
         try:
-            return await asyncio.wait_for(future, timeout)
-        except asyncio.exceptions.TimeoutError:
+            return await wait_for(future, timeout)
+        except TimeoutError:
             self.logger.error(
                 "Hit timeout of %d waiting for %s, %d",
                 timeout,
